# ==================================================================================================
# --- Imports
# ==================================================================================================

# Import third-party packages
import dash_mantine_components as dmc
from dash_iconify import DashIconify


# ==================================================================================================
# --- Header layout
# ==================================================================================================
def return_header_layout():
    """
    Returns the layout for the header of the simulation dashboard.

    Returns:
        dmc.Header: The header layout for the simulation dashboard.
    """

    def create_header_link(icon, href, size=22, color="cyan"):
        return dmc.Anchor(
            dmc.ThemeIcon(
                DashIconify(
                    icon=icon,
                    width=size,
                ),
                variant="outline",
                radius=30,
                size=36,
                color=color,
            ),
            href=href,
            target="_blank",
        )

    header = dmc.Header(
        height=70,
        fixed=True,
        px=20,
        children=[
            dmc.Stack(
                justify="center",
                align="center",
                style={"height": 70, "width": "100%"},
                children=dmc.Group(
                    position="apart",
                    style={"width": "100%"},
                    children=[
                        dmc.Text(
                            "ColBoard",
                            size=25,
                            color="cyan",
                            weight="bold",
                        ),
                        dmc.SegmentedControl(
                            id="tab-titles",
                            value="display-configuration",
                            data=[
                                {
                                    "value": "display-configuration",
                                    "label": "Configuration",
                                },
                                {"value": "display-knobs", "label": "Knobs"},
                                {"value": "display-twiss", "label": "Twiss"},
                                {"value": "display-scheme", "label": "Scheme"},
                                {"value": "display-sanity", "label": "Sanity checks"},
                                {"value": "display-separation", "label": "Separation"},
                                {
                                    "value": "display-3D-separation",
                                    "label": "3D separation",
                                },
                                {"value": "display-footprint", "label": "Footprint"},
                                {"value": "display-optics", "label": "Optics"},
                                {"value": "display-survey", "label": "Survey"},
                            ],
                            # color="cyan",
                            # style={"margin-right": "5%"},
                            fullWidth=True,
                            size="md",
                            color="cyan",
                        ),
                        dmc.Group(
                            position="right",
                            # spacing="xl",
                            children=[
                                create_header_link(
                                    "radix-icons:github-logo",
                                    "https://github.com/ColasDroin/simulation-dashboard",
                                ),
                            ],
                            mb=5,
                        ),
                    ],
                ),
            ),
        ],
    )
    return header



import json

import numpy as np


class FillingPatternSingleBeam(object):

    def __init__(self, pattern, ring_length_slots = 3564,
            min_MKI_slots = 31, min_MKP_slots = 7, agap_first_slot = 3443):

        agap_length = ring_length_slots - agap_first_slot
        pattern = np.int_(np.array(pattern))

        assert(len(pattern)<=ring_length_slots)

        # Add empty slots at the end if needed
        pattern = np.array(list(pattern) + (ring_length_slots-len(pattern))*[0])

        # Identify trains
        ddd = np.diff(pattern)
        start_trains = np.where(ddd==1)[0]+1
        end_trains = np.where(ddd==-1)[0]+1
        if end_trains[0]<start_trains[0]:
            start_trains = np.array([0] + list(start_trains))

        # Remove gaps that need to be generated in the PS
        tobekept = (start_trains[1:] - end_trains[:-1]) >= min_MKP_slots
        start_trains = np.array([start_trains[0]] + list(start_trains[1:][tobekept]))
        end_trains =  np.array(list(end_trains[:-1][tobekept]) + [end_trains[-1]])

        # Lenghts of trains and gaps
        train_lengths = end_trains - start_trains
        gap_lengths =  start_trains[1:] - end_trains[:-1]
        train_nbunches = np.array([np.sum(pattern[start_trains[ii]:end_trains[ii]]) for ii in \
                            range(len(start_trains))])

        i_train_first = np.array([0] + list(np.where(gap_lengths>=min_MKI_slots)[0]+1))
        inj_slots = start_trains[i_train_first]
        n_injections = len(inj_slots)

        #Find pattern for individual injections
        inj_patterns = []
        inj_compositions = []
        injection_number = 0*pattern - 1
        slot_within_injection = 0*pattern - 1
        for ii, _ in enumerate(i_train_first):

            i_start = inj_slots[ii]

            if ii+1 == n_injections:
                inj_composition = train_nbunches[i_train_first[ii]:]
                inj_pattern = pattern[i_start:]
            else:
                inj_composition = train_nbunches[i_train_first[ii]:i_train_first[ii+1]]
                i_end = inj_slots[ii+1]
                inj_pattern = pattern[i_start:i_end]

            # Cut away zeros at the end
            inj_last_filled = np.max(np.where(inj_pattern==1)[0])
            inj_pattern = inj_pattern[:inj_last_filled+1]
            inj_len = len(inj_pattern)
            injection_number[i_start:i_start+inj_len] = ii
            slot_within_injection[i_start:i_start+inj_len] = np.arange(inj_len)

            inj_patterns.append(inj_pattern)
            inj_compositions.append(inj_composition)

        # Identify injection types
        inj_pattern_types = []
        inj_composition_types = []
        for patt, comp in zip(inj_patterns, inj_compositions):
            found = False
            for patt_type in inj_pattern_types:
                if len(patt_type) == len(patt):
                    if np.sum(patt_type-patt)==0:
                        found = True
                        break
            if not found:
                inj_pattern_types.append(patt)
                inj_composition_types.append(comp)

        # Unused space
        needed_nslots = (n_injections-1)*min_MKI_slots \
                        + np.sum(list(map(len, inj_patterns))) + agap_length
        inefficiency_perc = 100*(1 - float(needed_nslots) / float(ring_length_slots))

        self.actual_MKP_slots = min(gap_lengths)
        self.actual_MKI_slots = min(gap_lengths[gap_lengths>min_MKP_slots])

        self.agap_length = agap_length

        self.pattern = pattern
        self.n_bunches = np.sum(pattern)

        self.train_lengths =  train_lengths
        self.gap_lengths = gap_lengths
        self.train_nbunches = train_nbunches

        self.inj_slots = inj_slots
        self.n_injections = n_injections

        self.inj_patterns = inj_patterns
        self.inj_compositions = inj_compositions
        self.inj_nbun = np.array(list(map(np.sum, inj_compositions)))

        self.inj_pattern_types = inj_pattern_types
        self.inj_composition_types = inj_composition_types
        self.inj_nbun_types = np.array(list(map(np.sum, inj_composition_types)))

        self.n_unused_slots = ring_length_slots - needed_nslots
        self.inefficiency_perc = inefficiency_perc

        self.injection_number = injection_number
        self.slot_within_injection = slot_within_injection

        self.ring_length_slots = ring_length_slots
        self.min_MKI_slots = min_MKI_slots
        self.min_MKP_slots = min_MKP_slots
        self.agap_first_slot = agap_first_slot

    def belongs_to_group(self, slots_within_injection=None,
            n_bunches_in_injection=None):

        mask = self.pattern>0

        # Check for nbun in injection
        if n_bunches_in_injection is not None:
            nbun_my_inj = np.take(self.inj_nbun, self.injection_number)
            nbun_my_inj[self.injection_number==-1]=-1
            mask = np.logical_and(mask, nbun_my_inj == n_bunches_in_injection)

        if slots_within_injection is not None:
            mask = np.logical_and(mask, np.array(
                [(ss in slots_within_injection) for ss in self.slot_within_injection]))

        return mask


class FillingPattern(object):

    @classmethod
    def from_json(cls, fname):
        with open(fname, 'r') as fid:
            data = json.load(fid)
        patt = cls(pattern_b1 = data['beam1'],
                   pattern_b2 = data['beam2'],
                   scheme_name=fname.split('.json')[0])
        return patt

    @classmethod
    def from_csv(cls, fname):

        with open(fname, 'r') as fid:
            lines = fid.readlines()

        patt_dict = {ll.split(',', 1)[0]: ll.split(',', 1)[1].replace('\n', '') for ll in lines}

        for bb in [1, 2]:
            # Build slot array
            patt_arr = np.array(3564*[False])
            filled_buckets = np.fromstring(patt_dict[f'B{bb} filled buckets'],
                       sep=',', dtype=np.int)
            for fbk in filled_buckets:
                assert(np.mod(fbk-1, 10) == 0)
                slot = (fbk-1)//10
                patt_arr[slot] = True
            patt_dict[f'pattern_array_b{bb}'] = patt_arr

            # Get gaps in slots
            mki_gap_in_slots = int(np.round(
                float(patt_dict[f'B{bb} LHC injection kicker gap [ns]'])/25 + - 1))
            patt_dict[f'mki_gap_in_slots_b{bb}'] = mki_gap_in_slots
            mkp_gap_in_slots = int(np.round(
                float(patt_dict[f'B{bb} SPS injection kicker gap [ns]'])/25 + - 1))
            patt_dict[f'mkp_gap_in_slots_b{bb}'] = mkp_gap_in_slots
            abort_gap_in_slots = int(np.round(
                float(patt_dict[f'B{bb} abort gap [ns]'])/25 + - 1))
            patt_dict[f'abort_gap_in_slots_b{bb}'] = abort_gap_in_slots

        patt_dict['mki_min'] = np.min([patt_dict[f'mki_gap_in_slots_b{bb}'] for bb in [1, 2]])
        patt_dict['mkp_min']= np.min([patt_dict[f'mkp_gap_in_slots_b{bb}'] for bb in [1, 2]])
        patt_dict['agap_min'] = np.min([patt_dict[f'abort_gap_in_slots_b{bb}'] for bb in [1, 2]])
        patt_dict['agap_first_slot'] = 3564 - patt_dict['agap_min']

        patt = cls(pattern_b1=patt_dict['pattern_array_b1'],
            pattern_b2=patt_dict['pattern_array_b2'],
            ring_length_slots = 3564,
            min_MKI_slots=patt_dict['mki_min'],
            min_MKP_slots=patt_dict['mkp_min'],
            agap_first_slot=patt_dict['agap_first_slot'],
            scheme_name=patt_dict['Scheme name'])

        return patt

    def __init__(self, pattern_b1, pattern_b2, ring_length_slots = 3564,
            min_MKI_slots = 31, min_MKP_slots = 7, agap_first_slot = 3443,
            scheme_name='no_name'):

        self.scheme_name = scheme_name
        self.b1 = FillingPatternSingleBeam(pattern_b1, ring_length_slots,
            min_MKI_slots, min_MKP_slots, agap_first_slot)

        self.b2 = FillingPatternSingleBeam(pattern_b2, ring_length_slots,
            min_MKI_slots, min_MKP_slots, agap_first_slot)

        self.b1.beam_name = 'beam 1'
        self.b2.beam_name = 'beam 2'

        self.n_coll_ATLAS = np.sum(self.b1.pattern*self.b2.pattern)
        self.n_coll_CMS = self.n_coll_ATLAS
        self.n_coll_ALICE = np.sum(self.b2.pattern * \
                                np.roll(self.b1.pattern, 891))
        self.n_coll_LHCb = np.sum(self.b1.pattern * \
                                np.roll(self.b2.pattern, 894))

    def to_csv(self, fname='auto'):

        txtlines = []
        txtlines.append('Scheme name,' + self.scheme_name)
        txtlines.append(','.join(['B1 filled buckets'] +
                    ['%d'%(bb*10+1) for bb in np.where(self.b1.pattern)[0]]))
        txtlines.append(','.join(['B2 filled buckets'] +
                    ['%d'%(bb*10+1) for bb in np.where(self.b2.pattern)[0]]))
        txtlines.append(','.join(['B1 injection buckets'] +
                    ['%d'%(bb*10+1) for bb in self.b1.inj_slots]))
        txtlines.append(','.join(['B2 injection buckets'] +
                    ['%d'%(bb*10+1) for bb in self.b2.inj_slots]))
        txtlines.append(','.join(['B1 injection n. bunches'] +
                    ['%d'%bb for bb in self.b1.inj_nbun]))
        txtlines.append(','.join(['B2 injection n. bunches'] +
                    ['%d'%bb for bb in self.b2.inj_nbun]))
        txtlines.append('N. coll. ATLAS/CMS, %d'%(self.n_coll_ATLAS))
        txtlines.append('N. coll. LHCb, %d'%(self.n_coll_LHCb))
        txtlines.append('N. coll. ALICE, %d'%(self.n_coll_ALICE))
        txtlines.append('B1 n. bunches, %d'%(self.b1.n_bunches))
        txtlines.append('B2 n. bunches, %d'%(self.b2.n_bunches))
        txtlines.append('B1 n. injections, %d'%(self.b1.n_injections))
        txtlines.append('B2 n. injections, %d'%(self.b2.n_injections))
        txtlines.append(','.join(['B1 max. injection length [ns]',
                    '%d'%((max(map(len, self.b1.inj_pattern_types)))*25)]))
        txtlines.append(','.join(['B2 max. injection length [ns]',
                    '%d'%((max(map(len, self.b2.inj_pattern_types)))*25)]))
        txtlines.append(','.join(['B1 LHC injection kicker gap [ns]',
                    '%d'%((self.b1.actual_MKI_slots+1)*25)]))
        txtlines.append(','.join(['B2 LHC injection kicker gap [ns]',
                    '%d'%((self.b2.actual_MKI_slots+1)*25)]))
        txtlines.append(','.join(['B1 SPS injection kicker gap [ns]',
                    '%d'%((self.b1.actual_MKP_slots+1)*25)]))
        txtlines.append(','.join(['B2 SPS injection kicker gap [ns]',
                    '%d'%((self.b2.actual_MKP_slots+1)*25)]))
        txtlines.append(','.join(['B1 abort gap [ns]',
                    '%d'%((self.b1.agap_length+1)*25)]))
        txtlines.append(','.join(['B2 abort gap [ns]',
                    '%d'%((self.b2.agap_length+1)*25)]))

        if fname == 'auto':
            fnamecsv = self.scheme_name + '.csv'
        else:
            fnamecsv = fname

        with open(fnamecsv, 'w') as fcsv:
            fcsv.write('\n'.join(txtlines))

    def compute_beam_beam_schedule(self, n_lr_per_side):
        from . import bbFunctions
        print('Computing collision schedules...')
        self.b1.bb_schedule = bbFunctions.B1CollisionScheduleDF(
            self.b1.pattern, self.b2.pattern,  n_lr_per_side)
        print('Done Beam 1')
        self.b2.bb_schedule = bbFunctions.B2CollisionScheduleDF(
            self.b1.pattern, self.b2.pattern,  n_lr_per_side)
        print('Done Beam 2')

        for bb in [self.b1, self.b2]:
            for ee in ['ATLAS/CMS', 'ALICE', 'LHCB']:
                 bb.bb_schedule[f'collides in {ee}'] = ~(np.isnan(bb.bb_schedule[f'HO partner in {ee}']))




# ==================================================================================================
# --- Imports
# ==================================================================================================

# Import third-party packages
import dash_mantine_components as dmc
from dash import dcc, html

# Import local functions
from ..backend import plot


# ==================================================================================================
# --- Optics layout
# ==================================================================================================
def return_optics_layout(dic_with_bb):
    """
    Returns the layout for the optics page of the simulation dashboard.

    Args:
        dic_with_bb (dict): A dictionary of collider (including beam-beam) observables.

    Returns:
        html.Div: A Div representing the layout for the optics tab.
    """
    optics_layout = html.Div(
        children=[
            dcc.Loading(
                dcc.Graph(
                    id="LHC-2D-near-IP",
                    mathjax=True,
                    config={
                        "displayModeBar": False,
                        "scrollZoom": True,
                        "responsive": True,
                        "displaylogo": False,
                    },
                    figure=plot.return_plot_optics(
                        dic_with_bb["df_tw_b1"],
                        dic_with_bb["df_tw_b2"],
                        dic_with_bb["df_sv_b1"],
                        dic_with_bb["df_elements_corrected"],
                        empty=True,
                    ),
                    style={"height": "90vh", "width": "100%", "margin": "auto"},
                ),
                type="circle",
                color="cyan",
            ),
            dmc.NumberInput(
                label="Vertical zoom",
                description="Vertical zoom level",
                id="vertical-zoom-optics",
                value=0,
                min=-14,
                max=14,
                step=1,
                style={
                    "width": 150,
                    "position": "absolute",
                    "bottom": "3%",
                    "right": "3%",
                },
            ),
        ],
        # style={"height": "100vh", "width": "100%", "margin": "auto"},
    )
    return optics_layout



"""Thid module is used to produce resonance line in tune-tune plot. The code directly adapted from 
the work of P. Belanger. See https://github.com/pbelange/BBStudies
"""
# ==================================================================================================
# --- Imports
# ==================================================================================================

# Third-party packages
from fractions import Fraction

import numpy as np
import pandas as pd
import plotly.graph_objects as go


# ==================================================================================================
# --- Functions
# ==================================================================================================
def farey(order):
    # Create array of (value,numerator,denominator)
    allFracts = [(0, 0, 1)] + [
        (m / k,) + Fraction(f"{m}/{k}").as_integer_ratio()
        for k in range(1, order + 1)
        for m in range(1, k + 1)
    ]
    uniqueFraq = np.array(list(set(allFracts)))
    # sort by value (1st columns) and return
    return list(uniqueFraq[np.argsort(uniqueFraq[:, 0]), 1:].astype(int))


def resonance_df(order):
    """Returns a dataframe with the information of resonance lines up to a given order"""
    resonances = []
    nodes = farey(int(order))
    for node_i in nodes:
        h, k = node_i  # Node h/k on the axes
        for node_j in nodes:
            p, q = node_j
            b = float(k * p)  # Resonance line a*Qx + b*Qy = c (linked to p/q)
            if b > 0:
                a, c = float(q - k * p), float(p * h)

                # Resonance lines from a*Qx + b*Qy = c
                # Qy = c/b - a/b*Qx
                # Qy = c/b + a/b*Qx
                # Qx = c/b - a/b*Qy     -> Qy = -(Qx - c/b)*b/a      if a!=0 else Qx = c/b, Qy = [0,1]
                # Qx = c/b + a/b*Qy     -> Qy =  (Qx - c/b)*b/a      if a!=0 else Qx = c/b, Qy = [0,1]
                # Qx = c/b - a/b*(1-Qy) -> Qy =  (Qx - (c-a)/b)*b/a  if a!=0 else Qx = c/b, Qy = [0,1]
                # Qx = c/b + a/b*(1-Qy) -> Qy = -(Qx - (c+a)/b)*b/a  if a!=0 else Qx = c/b, Qy = [0,1]

                if a != 0:
                    slopes = [-a / b, a / b, -b / a, b / a, b / a, -b / a]
                    y0s = [c / b, c / b, c / a, -c / a, -(c - a) / a, (c + a) / a]
                    x0s = [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]
                else:
                    slopes = [-a / b, a / b, np.inf]
                    y0s = [c / b, c / b, np.nan]
                    x0s = [np.nan, np.nan, c / b]

                for slope, y0, x0 in zip(slopes, y0s, x0s):
                    # Create unique ID to later eliminate duplicates
                    if slope != np.inf:
                        slope_int = Fraction(str(slope)).limit_denominator(20).as_integer_ratio()
                        y0_int = Fraction(str(y0)).limit_denominator(20).as_integer_ratio()
                        ID = slope_int + y0_int
                    else:
                        ID = (np.inf, np.inf) + Fraction(str(x0)).limit_denominator(
                            20
                        ).as_integer_ratio()

                    resonances.append(
                        {"ID": ID, "Order": int(a + b), "slope": slope, "y0": y0, "x0": x0}
                    )

            if q == k and p == 1:
                break

    resonances = pd.DataFrame(resonances)
    resonances = resonances.drop_duplicates(subset="ID").reset_index(drop=True)
    return resonances


def _is_line_inside(a, b, x_range, y_range, tol=0):
    """Assumes y = ax+b and finds if line goes inside the ROI"""
    x1, x2 = np.sort(x_range)
    y1, y2 = np.sort(y_range)
    tol = np.abs(tol)

    xi1 = (y1 - b) / a if a != 0 else 0
    xi2 = (y2 - b) / a if a != 0 else 0

    yi1 = a * x1 + b
    yi2 = a * x2 + b

    return (
        (x1 - tol <= xi1 <= x2 + tol)
        or (x1 - tol <= xi2 <= x2 + tol)
        or (y1 - tol <= yi1 <= y2 + tol)
        or (y1 - tol <= yi2 <= y2 + tol)
    )


def _plot_resonance_lines(
    df, Qx_range, Qy_range, ROI_tol=1e-3, offset=[0, 0], alpha=0.1, color="black"
):
    """Plots all resonance lines contained in a dataframe, provided by BBStudies.Physics.Resonances.resonance_df
    -> Filters out the lines that do not enter the ROI defined by Qx_range,Qy_range"""

    l_traces = []
    for _, line in df.iterrows():
        # Non-vertical lines
        if line["slope"] != np.inf:
            # Skip if line not in ROI
            if not _is_line_inside(line["slope"], line["y0"], Qx_range, Qy_range, tol=ROI_tol):
                continue

            xVec = np.array(Qx_range)
            yVec = line["slope"] * xVec + line["y0"]

        # Vertical line
        else:
            # Skip if line not in ROI
            if not (Qx_range[0] <= line["x0"] <= Qx_range[1]):
                continue

            xVec = line["x0"] * np.ones(2)
            yVec = np.array(Qy_range)

        factor_order = line["Order"] ** 0.5

        # Plotting if in ROI
        l_traces.append(
            go.Scattergl(
                x=xVec + offset[0],
                y=yVec + offset[1],
                mode="lines",
                line=dict(color=color, width=6 / factor_order),
                opacity=alpha,
            )
        )
    return l_traces


def get_working_diagram(
    Qx_range=[0, 1], Qy_range=[0, 1], order=6, offset=[0, 0], alpha=0.1, color="black"
):
    if not isinstance(order, (list, type(np.array([])))):
        # Regular, full working diagram
        resonances = resonance_df(order)
        l_traces = _plot_resonance_lines(
            resonances, Qx_range, Qy_range, ROI_tol=1e-3, offset=offset, alpha=alpha, color=color
        )
    else:
        l_traces = []
        # Selected resonances
        resonances = resonance_df(np.max(order))
        for _ord in order:
            l_traces.extend(
                _plot_resonance_lines(
                    resonances[resonances["Order"] == _ord],
                    Qx_range,
                    Qy_range,
                    ROI_tol=1e-3,
                    offset=offset,
                    alpha=alpha,
                    color=color,
                )
            )

    return l_traces


if __name__ == "__main__":
    # Plot example of resonance lines
    Qx_0, Qy_0 = 0.31, 0.32
    window = 0.05
    Qx_lim = [Qx_0 - 3 * window / 4, Qx_0 + window / 4]
    Qy_lim = [Qy_0 - 3 * window / 4, Qy_0 + window / 4]
    l_traces = get_working_diagram(order=12, Qx_range=Qx_lim, Qy_range=Qy_lim)
    fig = go.Figure(data=l_traces)
    fig.update_layout(
        xaxis=dict(range=Qx_lim, constrain="domain"),
        yaxis=dict(range=Qy_lim, scaleanchor="x", scaleratio=1, constrain="domain"),
        width=800,
        height=800,
    )
    fig.show()



from matplotlib import rc, rcdefaults, RcParams
import matplotlib
import pylab as pl
from colorsys import hsv_to_rgb
import matplotlib.pyplot as plt
import numpy as np


def mystyle(fontsz=16):
    rcdefaults()
    version = matplotlib.__version__.split('.')[0]
    if version == '2':
        print('Reverting matplotlib look to v1.5')
        plt.rcParams['axes.autolimit_mode'] = 'round_numbers'
        plt.rcParams['axes.xmargin'] = 0
        plt.rcParams['axes.ymargin'] = 0
        plt.rcParams['xtick.direction'] = 'in'
        plt.rcParams['ytick.direction'] = 'in'
        plt.rcParams['xtick.top'] = True
        plt.rcParams['ytick.right'] = True
        plt.rcParams['legend.numpoints'] = 1
        plt.style.use('classic')



    font = {#'family' : 'normal',
            #'weight' : 'bold',
            'size'   : fontsz}
#   print fontsz
    rc('font', **font)

def mystyle_arial(fontsz=16, dist_tick_lab=10):

    mystyle(fontsz)
    rc('font',**{'family':'sans-serif','sans-serif':['arial'], 'size':fontsz})
    rc(('xtick.major','xtick.minor','ytick.major','ytick.minor'), pad=dist_tick_lab)

def sciy():
    pl.gca().ticklabel_format(style='sci', scilimits=(0,0),axis='y')

def scix():
    pl.gca().ticklabel_format(style='sci', scilimits=(0,0),axis='x')

def colorprog(i_prog, Nplots, v1 = .9, v2 = 1., cm='hsv'):
    if hasattr(Nplots, '__len__'):
        Nplots = len(Nplots)
    if cm == 'hsv':
        return hsv_to_rgb(float(i_prog)/float(Nplots), v1, v2)
    elif cm == 'rainbow':
        return [pl.cm.rainbow(k) for k in np.linspace(0, 1, Nplots)][i_prog]
    else:
        raise ValueError('What?!')

def comb_legend(sp1, sp2, *args, **kwargs):
    """
    Combine legends for twinx()ed subplots
    """
    lines, labels = sp1.get_legend_handles_labels()
    lines2, labels2 = sp2.get_legend_handles_labels()
    sp2.legend(lines + lines2, labels + labels2, *args, **kwargs)


def mystyle_2(fontsz=16, dist_tick_lab=10, figsize=(12,10)):
    rcdefaults()
    RcParams[u'axes.grid'] = True
    RcParams[u'axes.linewidth'] = 2.0
    RcParams[u'figure.facecolor'] = 'w'

    rc('font',**{'family':'sans-serif','sans-serif':['arial'], 'size':fontsz})
    rc(('xtick.major','xtick.minor','ytick.major','ytick.minor'), pad=dist_tick_lab)


def figure(title, figs=None, figsize=(12, 10), **kwargs):
    fig = plt.figure(figsize=figsize, **kwargs)
    fig.canvas.set_window_title(title)
    if figs != None:
        figs.append(fig)
    fig.patch.set_facecolor('w')
    plt.suptitle(title, fontsize=16)
    fig.subplots_adjust(left=0.07, right=0.80, wspace=0.5)
    return fig


"""
RcParams


RcParams({u'agg.path.chunksize': 0,
          u'animation.avconv_args': [],
          u'animation.avconv_path': u'avconv',
          u'animation.bitrate': -1,
          u'animation.codec': u'mpeg4',
          u'animation.convert_args': [],
          u'animation.convert_path': u'convert',
          u'animation.ffmpeg_args': [],
          u'animation.ffmpeg_path': u'ffmpeg',
          u'animation.frame_format': u'png',
          u'animation.html': u'none',
          u'animation.mencoder_args': [],
          u'animation.mencoder_path': u'mencoder',
          u'animation.writer': u'ffmpeg',
          u'axes.axisbelow': False,
          u'axes.edgecolor': u'k',
          u'axes.facecolor': u'w',
          u'axes.formatter.limits': [-7, 7],
          u'axes.formatter.use_locale': False,
          u'axes.formatter.use_mathtext': False,
          u'axes.formatter.useoffset': True,
          u'axes.grid': False,
          u'axes.grid.axis': u'both',
          u'axes.grid.which': u'major',
          u'axes.hold': True,
          u'axes.labelcolor': u'k',
          u'axes.labelpad': 5.0,
          u'axes.labelsize': u'medium',
          u'axes.labelweight': u'normal',
          u'axes.linewidth': 1.0,
          u'axes.prop_cycle': cycler(u'color', [u'b', u'g', u'r', u'c', u'm', u'y', u'k']),
          u'axes.spines.bottom': True,
          u'axes.spines.left': True,
          u'axes.spines.right': True,
          u'axes.spines.top': True,
          u'axes.titlesize': u'large',
          u'axes.titleweight': u'normal',
          u'axes.unicode_minus': True,
          u'axes.xmargin': 0.0,
          u'axes.ymargin': 0.0,
          u'axes3d.grid': True,
          u'backend': u'Qt4Agg',
          u'backend.qt4': u'PyQt4',
          u'backend.qt5': u'PyQt5',
          u'backend_fallback': True,
          u'boxplot.bootstrap': None,
          u'boxplot.boxprops.color': u'b',
          u'boxplot.boxprops.linestyle': u'-',
          u'boxplot.boxprops.linewidth': 1.0,
          u'boxplot.capprops.color': u'k',
          u'boxplot.capprops.linestyle': u'-',
          u'boxplot.capprops.linewidth': 1.0,
          u'boxplot.flierprops.color': u'b',
          u'boxplot.flierprops.linestyle': u'none',
          u'boxplot.flierprops.linewidth': 1.0,
          u'boxplot.flierprops.marker': u'+',
          u'boxplot.flierprops.markeredgecolor': u'k',
          u'boxplot.flierprops.markerfacecolor': u'b',
          u'boxplot.flierprops.markersize': 6.0,
          u'boxplot.meanline': False,
          u'boxplot.meanprops.color': u'r',
          u'boxplot.meanprops.linestyle': u'-',
          u'boxplot.meanprops.linewidth': 1.0,
          u'boxplot.medianprops.color': u'r',
          u'boxplot.medianprops.linestyle': u'-',
          u'boxplot.medianprops.linewidth': 1.0,
          u'boxplot.notch': False,
          u'boxplot.patchartist': False,
          u'boxplot.showbox': True,
          u'boxplot.showcaps': True,
          u'boxplot.showfliers': True,
          u'boxplot.showmeans': False,
          u'boxplot.vertical': True,
          u'boxplot.whiskerprops.color': u'b',
          u'boxplot.whiskerprops.linestyle': u'--',
          u'boxplot.whiskerprops.linewidth': 1.0,
          u'boxplot.whiskers': 1.5,
          u'contour.corner_mask': True,
          u'contour.negative_linestyle': u'dashed',
          u'datapath': u'/afs/cern.ch/user/p/pyhdtl/public/anaconda/lib/python2.7/site-packages/matplotlib/mpl-data',
          u'docstring.hardcopy': False,
          u'errorbar.capsize': 3.0,
          u'examples.directory': u'',
          u'figure.autolayout': False,
          u'figure.dpi': 80.0,
          u'figure.edgecolor': u'w',
          u'figure.facecolor': u'0.75',
          u'figure.figsize': [8.0, 6.0],
          u'figure.frameon': True,
          u'figure.max_open_warning': 20,
          u'figure.subplot.bottom': 0.1,
          u'figure.subplot.hspace': 0.2,
          u'figure.subplot.left': 0.125,
          u'figure.subplot.right': 0.9,
          u'figure.subplot.top': 0.9,
          u'figure.subplot.wspace': 0.2,
          u'figure.titlesize': u'medium',
          u'figure.titleweight': u'normal',
          u'font.cursive': [u'Apple Chancery',
                            u'Textile',
                            u'Zapf Chancery',
                            u'Sand',
                            u'Script MT',
                            u'Felipa',
                            u'cursive'],
          u'font.family': [u'sans-serif'],
          u'font.fantasy': [u'Comic Sans MS',
                            u'Chicago',
                            u'Charcoal',
                            u'ImpactWestern',
                            u'Humor Sans',
                            u'fantasy'],
          u'font.monospace': [u'Bitstream Vera Sans Mono',
                              u'DejaVu Sans Mono',
                              u'Andale Mono',
                              u'Nimbus Mono L',
                              u'Courier New',
                              u'Courier',
                              u'Fixed',
                              u'Terminal',
                              u'monospace'],
          u'font.sans-serif': [u'Bitstream Vera Sans',
                               u'DejaVu Sans',
                               u'Lucida Grande',
                               u'Verdana',
                               u'Geneva',
                               u'Lucid',
                               u'Arial',
                               u'Helvetica',
                               u'Avant Garde',
                               u'sans-serif'],
          u'font.serif': [u'Bitstream Vera Serif',
                          u'DejaVu Serif',
                          u'New Century Schoolbook',
                          u'Century Schoolbook L',
                          u'Utopia',
                          u'ITC Bookman',
                          u'Bookman',
                          u'Nimbus Roman No9 L',
                          u'Times New Roman',
                          u'Times',
                          u'Palatino',
                          u'Charter',
                          u'serif'],
          u'font.size': 12.0,
          u'font.stretch': u'normal',
          u'font.style': u'normal',
          u'font.variant': u'normal',
          u'font.weight': u'normal',
          u'grid.alpha': 1.0,
          u'grid.color': u'k',
          u'grid.linestyle': u':',
          u'grid.linewidth': 0.5,
          u'image.aspect': u'equal',
          u'image.cmap': u'jet',
          u'image.composite_image': True,
          u'image.interpolation': u'bilinear',
          u'image.lut': 256,
          u'image.origin': u'upper',
          u'image.resample': False,
          u'interactive': True,
          u'keymap.all_axes': [u'a'],
          u'keymap.back': [u'left', u'c', u'backspace'],
          u'keymap.forward': [u'right', u'v'],
          u'keymap.fullscreen': [u'f', u'ctrl+f'],
          u'keymap.grid': [u'g'],
          u'keymap.home': [u'h', u'r', u'home'],
          u'keymap.pan': [u'p'],
          u'keymap.quit': [u'ctrl+w', u'cmd+w'],
          u'keymap.save': [u's', u'ctrl+s'],
          u'keymap.xscale': [u'k', u'L'],
          u'keymap.yscale': [u'l'],
          u'keymap.zoom': [u'o'],
          u'legend.borderaxespad': 0.5,
          u'legend.borderpad': 0.4,
          u'legend.columnspacing': 2.0,
          u'legend.edgecolor': u'inherit',
          u'legend.facecolor': u'inherit',
          u'legend.fancybox': False,
          u'legend.fontsize': u'large',
          u'legend.framealpha': None,
          u'legend.frameon': True,
          u'legend.handleheight': 0.7,
          u'legend.handlelength': 2.0,
          u'legend.handletextpad': 0.8,
          u'legend.isaxes': True,
          u'legend.labelspacing': 0.5,
          u'legend.loc': u'upper right',
          u'legend.markerscale': 1.0,
          u'legend.numpoints': 2,
          u'legend.scatterpoints': 3,
          u'legend.shadow': False,
          u'lines.antialiased': True,
          u'lines.color': u'b',
          u'lines.dash_capstyle': u'butt',
          u'lines.dash_joinstyle': u'round',
          u'lines.linestyle': u'-',
          u'lines.linewidth': 1.0,
          u'lines.marker': u'None',
          u'lines.markeredgewidth': 0.5,
          u'lines.markersize': 6.0,
          u'lines.solid_capstyle': u'projecting',
          u'lines.solid_joinstyle': u'round',
          u'markers.fillstyle': u'full',
          u'mathtext.bf': u'serif:bold',
          u'mathtext.cal': u'cursive',
          u'mathtext.default': u'it',
          u'mathtext.fallback_to_cm': True,
          u'mathtext.fontset': u'cm',
          u'mathtext.it': u'serif:italic',
          u'mathtext.rm': u'serif',
          u'mathtext.sf': u'sans\\-serif',
          u'mathtext.tt': u'monospace',
          u'nbagg.transparent': True,
          u'patch.antialiased': True,
          u'patch.edgecolor': u'k',
          u'patch.facecolor': u'b',
          u'patch.linewidth': 1.0,
          u'path.effects': [],
          u'path.simplify': True,
          u'path.simplify_threshold': 0.1111111111111111,
          u'path.sketch': None,
          u'path.snap': True,
          u'pdf.compression': 6,
          u'pdf.fonttype': 3,
          u'pdf.inheritcolor': False,
          u'pdf.use14corefonts': False,
          u'pgf.debug': False,
          u'pgf.preamble': [],
          u'pgf.rcfonts': True,
          u'pgf.texsystem': u'xelatex',
          u'plugins.directory': u'.matplotlib_plugins',
          u'polaraxes.grid': True,
          u'ps.distiller.res': 6000,
          u'ps.fonttype': 3,
          u'ps.papersize': u'letter',
          u'ps.useafm': False,
          u'ps.usedistiller': False,
          u'savefig.bbox': None,
          u'savefig.directory': u'~',
          u'savefig.dpi': 100.0,
          u'savefig.edgecolor': u'w',
          u'savefig.facecolor': u'w',
          u'savefig.format': u'png',
          u'savefig.frameon': True,
          u'savefig.jpeg_quality': 95,
          u'savefig.orientation': u'portrait',
          u'savefig.pad_inches': 0.1,
          u'savefig.transparent': False,
          u'svg.fonttype': u'path',
          u'svg.image_inline': True,
          u'svg.image_noscale': False,
          u'text.antialiased': True,
          u'text.color': u'k',
          u'text.dvipnghack': None,
          u'text.hinting': u'auto',
          u'text.hinting_factor': 8,
          u'text.latex.preamble': [],
          u'text.latex.preview': False,
          u'text.latex.unicode': False,
          u'text.usetex': False,
          u'timezone': u'UTC',
          u'tk.window_focus': False,
          u'toolbar': u'toolbar2',
          u'verbose.fileo': u'sys.stdout',
          u'verbose.level': u'silent',
          u'webagg.open_in_browser': True,
          u'webagg.port': 8988,
          u'webagg.port_retries': 50,
          u'xtick.color': u'k',
          u'xtick.direction': u'in',
          u'xtick.labelsize': u'medium',
          u'xtick.major.pad': 4.0,
          u'xtick.major.size': 4.0,
          u'xtick.major.width': 0.5,
          u'xtick.minor.pad': 4.0,
          u'xtick.minor.size': 2.0,
          u'xtick.minor.visible': False,
          u'xtick.minor.width': 0.5,
          u'ytick.color': u'k',
          u'ytick.direction': u'in',
          u'ytick.labelsize': u'medium',
          u'ytick.major.pad': 4.0,
          u'ytick.major.size': 4.0,
          u'ytick.major.width': 0.5,
          u'ytick.minor.pad': 4.0,
          u'ytick.minor.size': 2.0,
          u'ytick.minor.visible': False,
          u'ytick.minor.width': 0.5})
"""



# ==================================================================================================
# --- Imports
# ==================================================================================================
import dash_mantine_components as dmc


# ==================================================================================================
# --- Configuration layout
# ==================================================================================================
def return_knobs_layout(dic_knob_str):
    """
    Returns a Dash layout for displaying knob information.

    Args:
        dic_knob_str (str): A dictionnary containing the knob names (key) and information (values).

    Returns:
        dash.development.base_component.Component: A Dash layout component.
    """
    l_knobs = list(dic_knob_str.keys())
    if dic_knob_str is not None:
        configuration_layout = dmc.Stack(
            children=[
                dmc.Center(
                    dmc.Select(
                        id="select-knob_info",
                        data=l_knobs,
                        searchable=True,
                        nothingFound="No options found",
                        style={"width": 400},
                        value=l_knobs[0],
                    )
                ),
                dmc.Center(
                    children=[
                        dmc.Prism(
                            id="prism-knob_info",
                            language="python",
                            children="",
                            style={
                                "height": "90vh",
                                "overflowY": "auto",
                                "width": "80%",
                            },
                        ),
                    ],
                ),
            ],
        )

    else:
        configuration_layout = dmc.Center(
            dmc.Alert(
                "Configuration can't be displayed as no configuration file was provided.",
                title="No configuration !",
                style={"margin": "auto"},
            ),
            style={"width": "70%", "margin": "auto"},
        )

    return configuration_layout



# ==================================================================================================
# --- Imports to namespace
# ==================================================================================================
from .configuration import return_configuration_layout
from .knob import return_knobs_layout
from .filling import return_filling_scheme_layout
from .footprint import return_footprint_layout
from .header import return_header_layout
from .main import return_app_layout
from .optics import return_optics_layout
from .sanity import return_sanity_layout
from .separation import return_separation_layout
from .separation_3D import return_3D_separation_layout
from .survey import return_survey_layout
from .tables import return_tables_layout



import numpy as np 
from . import dotdict
import pandas as pd
dotdict=dotdict.dotdict


def computeBBMatrix(numberOfLRToConsider):
        """
        It returns a beam-beam matrix, that is a representation of the beam-beam encounters scheduled for a filled machine (3564 bunches).
        To obtain the BB encounters scheduled of the bunch N of B1 (for the machine totally full) you have to consider the N-row (e.g., BBMatrix[N,:]).
        To obtain the BB encounters scheduled of the bunch N of B2 (for the machine totally full) you have to consider the N-column (e.g., BBMatrix[:,N]).
                
        The numberOfLRToConsider represents the long-range number at the right and left of each IR (the total number of LR per IR is therefore 2 x numberOfLRToConsider).
        numberOfLRToConsider can be a scalar of an array of 3 scalars (longer arrays will be truncated to the length of 3).
        In case numberOfLRToConsider is as array the first scalar represent the long-range number at the right and left of ATLAS/CMS. 
        The second and third element of the array will be the number of long range encounters (at the right and left) in IR2 and IR8, respectively.
        
        The matrix element has a value 1, 2, 5 and 8 when there is a HO  in IP1, 2, 5 and 8, respectively.
        The matrix element has a value 10, 20, 50 and  80 when there is a LR respectively in IR1, 2, 5 and 8, respectively.
        
        It assumes that the positions of the IPs and the convention of the B1/B2 bunch numbering are such that:
        
        1. B1 Bunch 0 meets B2 Bunch 0 in IP1 and 5.
        2. B1 Bunch 0 meets B2 Bunch 891 in IP2.
        3. B1 Bunch 0 meets B2 Bunch 2670 in IP8.
        4. B2 Bunch 0 meets B1 Bunch 0 in IP1 and 5.
        5. B2 Bunch 0 meets B1 Bunch 2673 in IP2.
        6. B2 Bunch 0 meets B1 Bunch 894 in IP8.

        === EXAMPLE 1 ===
        myMatrix=computeBBMatrix(numberOfLRToConsider=20)
        #in this case the total number of LR will be 160: 40 in IR1/5, 40 in IR2, and 40 in IR8.

        === EXAMPLE 2 ===
        myMatrix=computeBBMatrix(numberOfLRToConsider=[20,15,16])
        #in this case the total number of LR will be 142: 40 in IR1/5, 30 in IR2, and 32 in IR8.
        """ 
        availableBunchSlot=3564
        BBMatrixLHC =np.zeros([availableBunchSlot,availableBunchSlot]);

        if isinstance(numberOfLRToConsider,int):
            numberOfLRToConsiderATLASCMS=numberOfLRToConsider
            numberOfLRToConsiderALICE=numberOfLRToConsider
            numberOfLRToConsiderLHCB=numberOfLRToConsider
        else:
            numberOfLRToConsiderATLASCMS=numberOfLRToConsider[0]
            numberOfLRToConsiderALICE=numberOfLRToConsider[1]
            numberOfLRToConsiderLHCB=numberOfLRToConsider[2]

        # HO in IP1 and IP5
        index=np.arange(availableBunchSlot)
        BBMatrixLHC[index,index]=1

        # BBLR in IP1 and IP5
        for i in range(1,numberOfLRToConsiderATLASCMS+1):
            index=np.arange(availableBunchSlot-i)
            BBMatrixLHC[index,index+i]=10
            BBMatrixLHC[index,index-i]=10
            BBMatrixLHC[index+i,index]=10
            BBMatrixLHC[index-i,index]=10

        # HO in IP2
        IP2slot=int(availableBunchSlot/4)
        index=np.arange(availableBunchSlot-IP2slot)
        BBMatrixLHC[index,index+IP2slot]=2
        index=np.arange(availableBunchSlot-IP2slot,availableBunchSlot)
        BBMatrixLHC[index,index-(availableBunchSlot-IP2slot)]=2

        # BBLR in IP2
        for i in range(1,numberOfLRToConsiderALICE+1):
            index=np.arange(availableBunchSlot-IP2slot-i)
            BBMatrixLHC[index,index+IP2slot+i]=20
            BBMatrixLHC[index+i,index+IP2slot]=20
            BBMatrixLHC[index,index+IP2slot-i]=20
            BBMatrixLHC[index-i,index+IP2slot]=20
            index=np.arange(availableBunchSlot-IP2slot,availableBunchSlot-i)
            BBMatrixLHC[index,index-(availableBunchSlot-IP2slot)+i]=20
            BBMatrixLHC[index+i,index-(availableBunchSlot-IP2slot)]=20
            BBMatrixLHC[index,index-(availableBunchSlot-IP2slot)-i]=20
            BBMatrixLHC[index-i,index-(availableBunchSlot-IP2slot)]=20

        # HO in IP8
        IP8slot=int(availableBunchSlot/4*3-3)
        index=np.arange(availableBunchSlot-IP8slot)
        BBMatrixLHC[index,index+IP8slot]=8
        index=np.arange(availableBunchSlot-IP8slot,availableBunchSlot)
        BBMatrixLHC[index,index-(availableBunchSlot-IP8slot)]=8

        # BBLR in IP8
        for i in range(1,numberOfLRToConsiderLHCB+1):
            index=np.arange(availableBunchSlot-IP8slot-i)
            BBMatrixLHC[index,index+IP8slot+i]=80
            BBMatrixLHC[index+i,index+IP8slot]=80
            BBMatrixLHC[index,index+IP8slot-i]=80
            BBMatrixLHC[index-i,index+IP8slot]=80
            index=np.arange(availableBunchSlot-IP8slot,availableBunchSlot-i)
            BBMatrixLHC[index,index-(availableBunchSlot-IP8slot)+i]=80
            BBMatrixLHC[index+i,index-(availableBunchSlot-IP8slot)]=80
            BBMatrixLHC[index,index-(availableBunchSlot-IP8slot)-i]=80
            BBMatrixLHC[index-i,index-(availableBunchSlot-IP8slot)]=80

        return BBMatrixLHC;

def _bunch_BB_pattern(Bunch,BBMatrixLHC):
    '''
    It returns the beam-beam pattern of a bunch of B1 and B2 [adimensional array of integer] in ALICE, ATLAS, CMS, LHCB.
    
    - Bunch [adimensional integer]: the bunch number in B1 and B2 to consider.
    - BBMatrixLHC [adimensional integer matrix]: the beam-beam matrix to consider (see bbFunctions.computeBBMatrix?).
    The returned array is ordered with respect to the positive direction of B1 (clockwise in LHC). 
    WARNING: the bunch number is defined wrt the negative direction of each beam.
    
    Conventions :
    
    ...=>B1 bunch 1 => B1 bunch 0 => |[IP]| <= B2 bunch 0 <= B2 bunch 1 <=...
    
    This means that B1 bunch 0 will meet B2 bunch 1 on the positive side of the IP and B2 bunch 0 will meet B1 bunch 0 on the negative side of IP1 (positive/negative wrt the B1).
    
    '''
    
    #### BEAM 1 ####
    BBVector=BBMatrixLHC[Bunch,:]
    
    numberOfLRToConsider=len(np.where(BBMatrixLHC[Bunch,:]==10)[0])/2
    HO_in_IP=BBVector==1
    LR_in_IP=BBVector==10
    aux=np.where((LR_in_IP) | (HO_in_IP))[0]
    np.where(aux==Bunch)[0]
    B1=np.roll(aux,  numberOfLRToConsider-np.where(aux==np.where(HO_in_IP)[0][0])[0])
    resultsB1=dotdict({'atATLAS':B1,
         'atCMS':B1})
    
    HO_in_IP=BBVector==2
    LR_in_IP=BBVector==20
    numberOfLRToConsider=len(np.where(BBMatrixLHC[Bunch,:]==20)[0])/2
    aux=np.where((LR_in_IP) | (HO_in_IP))[0]
    np.where(aux==Bunch)[0]
    B1=np.roll(aux, numberOfLRToConsider-np.where(aux==np.where(HO_in_IP)[0][0])[0])
    resultsB1.update({'atALICE':B1})
    
    HO_in_IP=BBVector==8
    LR_in_IP=BBVector==80
    numberOfLRToConsider=len(np.where(BBMatrixLHC[Bunch,:]==80)[0])/2
    aux=np.where((LR_in_IP) | (HO_in_IP))[0]
    np.where(aux==Bunch)[0]
    B1=np.roll(aux, numberOfLRToConsider-np.where(aux==np.where(HO_in_IP)[0][0])[0])
    resultsB1.update({'atLHCB':B1})
    
    #### BEAM 2 ####
    BBVector=BBMatrixLHC[:,Bunch]
    numberOfLRToConsider=len(np.where(BBMatrixLHC[Bunch,:]==10)[0])/2
    HO_in_IP=BBVector==1
    LR_in_IP=BBVector==10
    aux=np.where((LR_in_IP) | (HO_in_IP))[0]
    np.where(aux==Bunch)[0]
    B2=np.roll(aux,  numberOfLRToConsider-np.where(aux==np.where(HO_in_IP)[0][0])[0])
    resultsB2=dotdict({'atATLAS':B2[::-1], 
         'atCMS':B2[::-1]}) # To note the inverstion of the direction
   
    HO_in_IP=BBVector==2
    LR_in_IP=BBVector==20
    numberOfLRToConsider=len(np.where(BBMatrixLHC[Bunch,:]==20)[0])/2
    aux=np.where((LR_in_IP) | (HO_in_IP))[0]
    np.where(aux==Bunch)[0]
    B2=np.roll(aux, numberOfLRToConsider-np.where(aux==np.where(HO_in_IP)[0][0])[0])
    resultsB2.update({'atALICE':B2[::-1]})
    
    HO_in_IP=BBVector==8
    LR_in_IP=BBVector==80
    numberOfLRToConsider=len(np.where(BBMatrixLHC[Bunch,:]==80)[0])/2
    aux=np.where((LR_in_IP) | (HO_in_IP))[0]
    np.where(aux==Bunch)[0]
    B2=np.roll(aux, numberOfLRToConsider-np.where(aux==np.where(HO_in_IP)[0][0])[0])
    resultsB2.update({'atLHCB':B2[::-1]})
    # The encounters seen by  B1/2 are in increasing/decreasing order
    return dotdict({'atB1':resultsB1, 'atB2':resultsB2})

def BBEncounterSchedule(B1_fillingScheme,B2_fillingScheme,BBMatrixLHC):
    """
    It returns a dictionary structure with the BB encounters of B1 and B2 taking into account the filling schemes.
    - B1_fillingScheme [adimensional integer array]: the B1 filling scheme.
    - B2_fillingScheme [adimensional integer array]: the B2 filling scheme.
    The dictionary structure has the following hierarchy:
    - BEAM >> BUNCH >> EXPERIMENT >> ENCOUNTERS
    - BEAM >> BUNCH >> EXPERIMENT >> POSITIONS
    All the positions are referred to the positive direction of B1 (clockwise in LHC).
    WARNING: the bunch number is defined wrt the negative direction of each beam.
    
    === EXAMPLE 1 ===
    from cl2pd import bbFunctions
    from cl2pd import importData
   
    np=importData.np  
    BBMatrixLHC=bbFunctions.computeBBMatrix(numberOfLRToConsider=25)
    B1_bunches=np.array([0,1,2])
    B2_bunches=np.array([0,1,2])
    results=beam_BB_pattern(B1_bunches, B2_bunches, BBMatrixLHC)
    #or more realistically
    fillingSchemeDF=importData.LHCCals2pd(['LHC.BCTFR.A6R4.B%:BUNCH_FILL_PATTERN'],6972, ['FLATTOP'],flag='next')
    B1_bunches = fillingSchemeDF['LHC.BCTFR.A6R4.B1:BUNCH_FILL_PATTERN'].iloc[0]
    B2_bunches = fillingSchemeDF['LHC.BCTFR.A6R4.B2:BUNCH_FILL_PATTERN'].iloc[0]
    results=bbFunctions.BBEncounterSchedule(np.where(B1_bunches)[0], np.where(B2_bunches)[0], BBMatrixLHC)
    """
    experiments=['atALICE','atATLAS','atCMS','atLHCB']

    #B1
    B1_BB_pattern=dotdict()
    for i in B1_fillingScheme:
        bunch_aux=dotdict()
        for j in experiments:
            results=_bunch_BB_pattern(i,BBMatrixLHC)
            B2=results['atB1'][j]
            aux=B2[np.in1d(B2,B2_fillingScheme)]
            myPosition=np.arange(-(len(B2)-1)/2,(len(B2)-1)/2+1)
            bunch_aux.update({j: {'atEncounters' : aux,'atPositions':myPosition[np.in1d(B2,B2_fillingScheme)]}})
            B1_BB_pattern.update({'at'+format(i,'04d'):bunch_aux})

    #B2
    B2_BB_pattern=dotdict()
    for i in B2_fillingScheme:
        bunch_aux=dotdict()
        for j in experiments:
            results=_bunch_BB_pattern(i,BBMatrixLHC)
            B1=results['atB2'][j]
            aux=B1[np.in1d(B1,B1_fillingScheme)]
            myPosition=np.arange(-(len(B1)-1)/2,(len(B1)-1)/2+1)
            bunch_aux.update({j: {'atEncounters' : aux,'atPositions':myPosition[np.in1d(B1,B1_fillingScheme)]}})
            B2_BB_pattern.update({'at'+format(i,'04d'):bunch_aux})

    beam_BB_pattern=dotdict({'atB1':B1_BB_pattern,'atB2':B2_BB_pattern})
    return beam_BB_pattern

def B1CollisionScheduleDF (B1_bunches, B2_bunches, numberOfLRToConsider):
    
    '''
    For given two series of booleans which represent bunches and a array that represent long range collisions, 
    this function returns dataframe related to their collisions from perspective of beam 1.
    
    ===EXAMPLE===
    fillingSchemeDF=importData.LHCFillsAggregation(['LHC.BCTFR.A6R4.B%:BUNCH_FILL_PATTERN'],6666, ['FLATTOP'],flag='next')
    B1_bunches = fillingSchemeDF['LHC.BCTFR.A6R4.B1:BUNCH_FILL_PATTERN'].iloc[0]
    B2_bunches = fillingSchemeDF['LHC.BCTFR.A6R4.B2:BUNCH_FILL_PATTERN'].iloc[0]
    B1CollisionScheduleDF(B1_bunches, B2_bunches, 25)
    
    '''
    bunch_value = 1.0
    # Transforming bunches in to boolean array
    B1_bunches = np.array(B1_bunches) == bunch_value
    B2_bunches = np.array(B2_bunches) == bunch_value
    
    # For debugging
    # pdb.set_trace()
    
    # Get indexes of Bunches
    B1_bunches_index = np.where(B1_bunches)[0]
    B2_bunches_index = np.where(B2_bunches)[0]
    
    if isinstance(numberOfLRToConsider, int):
        numberOfLRToConsider = [numberOfLRToConsider, numberOfLRToConsider, numberOfLRToConsider]
    
    B1df = pd.DataFrame() 
       
    for n in B1_bunches_index:
    
    
        # First check for collisions in ALICE

        # Formula for head on collision in ALICE is 
        # (n + 891) mod 3564 = m
        # where n is number of bunch in B1, and m is number of bunch in B2
        
        # Formula for head on collision in ATLAS/CMS is 
        # n = m
        # where n is number of bunch in B1, and m is number of bunch in B2
        
        # Formula for head on collision in LHCb is 
        # (n + 2670) mod 3564 = m
        # where n is number of bunch in B1, and m is number of bunch in B2

        head_on_names = ["HO partner in ALICE", "HO partner in ATLAS/CMS", "HO partner in LHCB"]
        secondary_names = ["# of LR in ALICE", "# of LR in ATLAS/CMS", "# of LR in LHCB"]
        encounters_names = ["BB partners in ALICE", "BB partners in ATLAS/CMS", "BB partners in LHCB"]
        positions_names = ["Positions in ALICE", "Positions in ATLAS/CMS", "Positions in LHCB"]
        
        colide_factor_list = [891, 0, 2670]
        number_of_bunches = 3564
        
        # i == 0 for ALICE
        # i == 1 for ATLAS and CMS
        # i == 2 for LHCB
        
        dictonary = {}
        
        for i in range(0,3):
        
            collide_factor = colide_factor_list[i]
            m = (n + collide_factor) % number_of_bunches
            
            #pdb.set_trace()
            # if this Bunch is true, than there is head on collision
            if B2_bunches[m]:
                head_on = m
            else:
                head_on = np.nan
            
            ## Check if beam 2 has bunches in range  m - numberOfLRToConsider to m + numberOfLRToConsider 
            ## Also have to check if bunches wrap around from 3563 to 0 or vice versa
            
            
            bunches_ineraction_temp = np.array([])
            encounters = np.array([])
            positions = np.array([])
            
            first_to_consider = m - numberOfLRToConsider[i]
            last_to_consider = m + numberOfLRToConsider[i] + 1
            
            numb_of_long_range = 0
                
            if first_to_consider < 0:
                bunches_ineraction_partial = np.where(B2_bunches[(number_of_bunches + first_to_consider):(number_of_bunches)])[0]
                                
                # This represents the absolute position of the bunches 
                encounters = np.append(encounters, number_of_bunches + first_to_consider + bunches_ineraction_partial )
                
                #This represents the relative position to the head-on bunch
                positions = np.append(positions, first_to_consider + bunches_ineraction_partial)
                
                #Set this varibale so later the normal syntax wihtout the wrap around checking can be used
                first_to_consider = 0
                
            if last_to_consider > number_of_bunches:
                bunches_ineraction_partial = np.where(B2_bunches[0:last_to_consider - number_of_bunches])[0]
                                
                # This represents the absolute position of the bunches 
                encounters = np.append(encounters, bunches_ineraction_partial )
                
                #This represents the relative position to the head-on bunch
                positions = np.append(positions, number_of_bunches - m + bunches_ineraction_partial)
                
                
                last_to_consider = number_of_bunches
            
            bunches_ineraction_partial = np.append(bunches_ineraction_temp, np.where(B2_bunches[first_to_consider:last_to_consider])[0])
                        
            # This represents the absolute position of the bunches 
            encounters = np.append(encounters, first_to_consider + bunches_ineraction_partial)
            
            #This represents the relative position to the head-on bunch
            positions = np.append(positions, bunches_ineraction_partial - (m - first_to_consider))
            
            
            # Substract head on collision from number of secondary collisions
            numb_of_long_range = len(positions) - int(B2_bunches[m])

            
            dictonary.update({head_on_names[i] : { n : head_on }, secondary_names[i] : { n : numb_of_long_range }, \
                             encounters_names[i] : { n : encounters },  positions_names[i] : { n : positions } })

        B1df = pd.concat([B1df,pd.DataFrame(dictonary)])


    return B1df

def B2CollisionScheduleDF (B1_bunches, B2_bunches, numberOfLRToConsider):
    
    '''
    For given two series of booleans which represent bunches and a array that represent long range collisions, 
    this function returns dataframe related to their collisions from perspective of beam 1.
    
    ===EXAMPLE===
    fillingSchemeDF=importData.LHCFillsAggregation(['LHC.BCTFR.A6R4.B%:BUNCH_FILL_PATTERN'],6666, ['FLATTOP'],flag='next')
    B1_bunches = fillingSchemeDF['LHC.BCTFR.A6R4.B1:BUNCH_FILL_PATTERN'].iloc[0]
    B2_bunches = fillingSchemeDF['LHC.BCTFR.A6R4.B2:BUNCH_FILL_PATTERN'].iloc[0]
    B2CollisionScheduleDF(B1_bunches, B2_bunches, 25)
    
    '''
    bunch_value = 1.0
    # Transforming bunches in to boolean array
    B1_bunches = np.array(B1_bunches) == bunch_value
    B2_bunches = np.array(B2_bunches) == bunch_value
    
    # For debugging
    # pdb.set_trace()
    
    # Get indexes of Bunches
    B1_bunches_index = np.where(B1_bunches)[0]
    B2_bunches_index = np.where(B2_bunches)[0]
    
    if isinstance(numberOfLRToConsider, int):
        numberOfLRToConsider = [numberOfLRToConsider, numberOfLRToConsider, numberOfLRToConsider]
    
    B2df = pd.DataFrame() 
       
    for n in B2_bunches_index:
    
    
        # First check for collisions in ALICE

        # Formula for head on collision in ALICE is 
        # (n + 891) mod 3564 = m
        # where n is number of bunch in B1, and m is number of bunch in B2
        
        # Formula for head on collision in ATLAS/CMS is 
        # n = m
        # where n is number of bunch in B1, and m is number of bunch in B2
        
        # Formula for head on collision in LHCb is 
        # (n + 2670) mod 3564 = m
        # where n is number of bunch in B1, and m is number of bunch in B2

        head_on_names = ["HO partner in ALICE", "HO partner in ATLAS/CMS", "HO partner in LHCB"]
        secondary_names = ["# of LR in ALICE", "# of LR in ATLAS/CMS", "# of LR in LHCB"]
        encounters_names = ["BB partners in ALICE", "BB partners in ATLAS/CMS", "BB partners in LHCB"]
        positions_names = ["Positions in ALICE", "Positions in ATLAS/CMS", "Positions in LHCB"]
        
        colide_factor_list = [891, 0, 2670]
        number_of_bunches = 3564
        
        # i == 0 for ALICE
        # i == 1 for ATLAS and CMS
        # i == 2 for LHCB
        
        dictonary = {}
        
        for i in range(0,3):
        
            collide_factor = colide_factor_list[i]
            m = (n - collide_factor) % number_of_bunches
            
            #pdb.set_trace()
            # if this Bunch is true, than there is head on collision
            if B1_bunches[m]:
                head_on = m
            else:
                head_on = np.nan
            
            ## Check if beam 2 has bunches in range  m - numberOfLRToConsider to m + numberOfLRToConsider 
            ## Also have to check if bunches wrap around from 3563 to 0 or vice versa
            
            
            bunches_ineraction_temp = np.array([])
            encounters = np.array([])
            positions = np.array([])
            
            first_to_consider = m - numberOfLRToConsider[i]
            last_to_consider = m + numberOfLRToConsider[i] + 1
            
            numb_of_long_range = 0
                
            if first_to_consider < 0:
                bunches_ineraction_partial = np.where(B1_bunches[(number_of_bunches + first_to_consider):(number_of_bunches)])[0]
                                
                # This represents the absolute position of the bunches 
                encounters = np.append(encounters, number_of_bunches + first_to_consider + bunches_ineraction_partial )
                
                #This represents the relative position to the head-on bunch
                positions = np.append(positions, first_to_consider + bunches_ineraction_partial)
                
                #Set this varibale so later the normal syntax wihtout the wrap around checking can be used
                first_to_consider = 0
                
            if last_to_consider > number_of_bunches:
                bunches_ineraction_partial = np.where(B1_bunches[0:last_to_consider - number_of_bunches])[0]
                                
                # This represents the absolute position of the bunches 
                encounters = np.append(encounters, bunches_ineraction_partial )
                
                #This represents the relative position to the head-on bunch
                positions = np.append(positions, number_of_bunches - m + bunches_ineraction_partial)
                
                
                last_to_consider = number_of_bunches
            
            bunches_ineraction_partial = np.append(bunches_ineraction_temp, np.where(B1_bunches[first_to_consider:last_to_consider])[0])
                        
            # This represents the absolute position of the bunches 
            encounters = np.append(encounters, first_to_consider + bunches_ineraction_partial)
            
            #This represents the relative position to the head-on bunch
            positions = np.append(positions, bunches_ineraction_partial - (m - first_to_consider))            
            
            # Substract head on collision from number of secondary collisions
            numb_of_long_range = len(positions) - int(B1_bunches[m])

            
            dictonary.update({head_on_names[i] : { n : head_on }, secondary_names[i] : { n : numb_of_long_range }, \
                             encounters_names[i] : { n : encounters },  positions_names[i] : { n : positions } })

        B2df = pd.concat([B2df,pd.DataFrame(dictonary)])


    return B2df

def _BeamFilling2str(booleanList, flagChars = ['e', 'b']):
    '''
    
    For a given boolean list, this function returns a string which represents the counting operation of boolean values. For example
    booleanList = [False, False, True, False, False, False] would return "2e1b3e"
    Character flags that represent true or false can be edited with flagChars variable
    
    b - bunch ~ True
    e - empty ~ False
    
    ===EXAMPLE===
    genericFillingPatern([True, True, True, False, False, False, True])
        
    '''

    
    #pdb.set_trace()
    
    counter = [0, 0]
    result = ''
    
    lengthList = len(booleanList)
    if lengthList == 0:
        raise Error()
    currentStatus = booleanList[0]
    
    for i in range(0, lengthList):
        
        if currentStatus == booleanList[i]:
            counter[int(currentStatus)] = counter[int(currentStatus)] + 1
            if i == lengthList - 1:
                result = result + str(counter[int(currentStatus)]) + flagChars[int(currentStatus)]    
        else:
            result = result + str(counter[int(currentStatus)]) + flagChars[int(currentStatus)]
            counter[int(currentStatus)] = 0
            
            currentStatus = not currentStatus
            counter[int(currentStatus)] = 1
            
            if i == lengthList - 1:
                result = result + str(counter[int(currentStatus)]) + flagChars[int(currentStatus)]
            
    return result

def BeamFilling2str (bunchFilling, flagChars = ['e', 'b']):
    
    '''
    
    Simple function that addapts the _BeamFilling2str for use in the cals2pd package
    
    Text from _BeamFilling2str:    
    For a given boolean list, this function returns a string which represents the counting operation of boolean values. For example
    booleanList = [False, False, True, False, False, False] would return "2e1b3e"
    Character flags that represent true or false can be edited with flagChars variable
    
    b - bunch ~ True
    e - empty ~ False
    
    ===EXAMPLE===
    fillingSchemeDF=importData.LHCFillsAggregation(['LHC.BCTFR.A6R4.B%:BUNCH_FILL_PATTERN'],6666, ['FLATTOP'],flag='next')
    BeamFilling2str(fillingSchemeDF['LHC.BCTFR.A6R4.B1:BUNCH_FILL_PATTERN'].iloc[0])
        
    '''
    
    bunchesBoolean = np.array(bunchFilling) == 1.0
    return _BeamFilling2str(bunchesBoolean, flagChars)
